import concurrent.futures

import numpy as np
import pandas as pd
from sentence_transformers import SentenceTransformer
from sklearn.feature_extraction.text import CountVectorizer, TfidfVectorizer
from tqdm.auto import tqdm

from sam_ml.config import setup_logger

logger = setup_logger(__name__)


class Embeddings_builder:
    """ Vectorizer Wrapper class """

    def __init__(self, vec: str = "count", **kwargs):
        """
        @param:
            vec:
                'count': CountVectorizer (default)
                'tfidf': TfidfVectorizer
                'bert': SentenceTransformer("quora-distilbert-multilingual")

            **kwargs:
                additional parameters for CountVectorizer or TfidfVectorizer
        """
        self.vec_type = vec
        self._grid: dict[str, list] = {} # for pipeline structure

        if vec == "bert":
            logger.debug("using quora-distilbert-multilingual model as vectorizer")
            self.vectorizer = SentenceTransformer("quora-distilbert-multilingual")

        elif vec == "count":
            logger.debug("using CountVectorizer as vectorizer")
            self.vectorizer = CountVectorizer(**kwargs)

        elif vec == "tfidf":
            logger.debug("using TfidfVectorizer as vectorizer")
            self.vectorizer = TfidfVectorizer(**kwargs)

        else:
            logger.error(f"the entered vectorizer '{vec}' cannot be used --> using CountVectorizer as vectorizer")
            self.vectorizer = CountVectorizer()
            self.vec_type = "count"

    def __repr__(self) -> str:
        vec_params: str = ""
        param_dict = self.get_params(False)
        for key in param_dict:
            if type(param_dict[key]) == str:
                vec_params += key+"='"+str(param_dict[key])+"', "
            else:
                vec_params += key+"="+str(param_dict[key])+", "
        return f"Embeddings_builder({vec_params})"

    @staticmethod
    def params() -> dict:
        """
        @return:
            possible values for the parameters
        """
        param = {"vec": ["bert", "count", "tfidf"]}
        return param

    def get_params(self, deep: bool = True):
        class_params = {"vec": self.vec_type}
        if self.vec_type != "bert":
            return class_params | self.vectorizer.get_params(deep)
        return class_params | {"model_name_or_path": "quora-distilbert-multilingual"}

    def set_params(self, **params):
        if self.vec_type == "bert":
            self.vectorizer = SentenceTransformer("quora-distilbert-multilingual", **params)
        else:
            self.vectorizer.set_params(**params)
        return self
    
    def create_parallel_bert_embeddings(self, content: list) -> list:
        logger.debug("Going to parallel process embedding creation")

        # Create a progress bar
        pbar = tqdm(total=len(content), desc="Bert Embeddings")

        # Define a new function that updates the progress bar after each embedding
        def get_embedding_and_update(text: str) -> list:
            pbar.update()
            return self.vectorizer.encode(text)
        
        # Parallel processing
        with concurrent.futures.ThreadPoolExecutor() as executor:
            content_embeddings = list(executor.map(get_embedding_and_update, content))

        # Close the progress bar
        pbar.close()

        return content_embeddings

    def vectorize(self, data: pd.Series, train_on: bool = True) -> pd.DataFrame:
        """
        @params:
            data: pandas Series
            train_on: shall the vectorizer fit before transform
        @return:
            pandas Dataframe with vectorized data
        """
        indices = data.index
        logger.debug("creating embeddings - started")
        if self.vec_type == "bert":
            message_embeddings = self.create_parallel_bert_embeddings(list(data))
            emb_ar = np.asarray(message_embeddings)

        else:
            if train_on:
                emb_ar = self.vectorizer.fit_transform(data).toarray()
            else:
                emb_ar = self.vectorizer.transform(data).toarray()

        emb_df = pd.DataFrame(emb_ar, index=indices).add_suffix("_"+data.name)
        logger.debug("creating embeddings - finished")

        return emb_df



from ConfigSpace import Categorical, ConfigurationSpace, Float
from sklearn.svm import SVC as svc

from .main_classifier import Classifier


class SVC(Classifier):
    """ SupportVectorClassifier Wrapper class """

    def __init__(
        self,
        model_name: str = "SupportVectorClassifier",
        kernel: str = "rbf",
        random_state: int = 42,
        **kwargs,
    ):
        """
        @param (important one):
            random_state: random_state for model
            verbose: logging (True/False)
            C: Inverse of regularization strength
            kernel: kernel type to be used in the algorithm
            gamma: Kernel coefficient for 'rbf', 'poly' and 'sigmoid'

            class_weight: set class_weight="balanced" to deal with imbalanced data
            probability: probability=True enables probability estimates for SVM algorithms

            cache_size: Specify the size of the kernel cache (in MB)
        """
        model_type = "SVC"
        model = svc(
            kernel=kernel,
            random_state=random_state,
            **kwargs,
        )
        grid = ConfigurationSpace(
            seed=42,
            space={
            "kernel": Categorical("kernel", ["rbf", "poly", "sigmoid"], default="rbf"),
            "gamma": Float("gamma", (0.0001, 1), log=True, default=0.001),
            "C": Float("C", (0.1, 1000), log=True, default=1),
            "probability": Categorical("probability", [True, False], default=False),
            })
        super().__init__(model, model_name, model_type, grid)

    def feature_importance(self):
        if self.model.kernel == "linear":
            super().feature_importance()
        else:
            print(f"feature importance is only available for a linear kernel. You are currently using: {self.model.kernel}")



import numpy as np
from ConfigSpace import (
    Categorical,
    ConfigurationSpace,
    EqualsCondition,
    Float,
    ForbiddenAndConjunction,
    ForbiddenEqualsClause,
    ForbiddenInClause,
)
from sklearn.linear_model import LogisticRegression

from .main_classifier import Classifier


class LR(Classifier):
    """ LogisticRegression Wrapper class """

    def __init__(
        self,
        model_name: str = "LogisticRegression",
        random_state: int = 42,
        **kwargs,
    ):
        """
        @param (important one):
            n_jobs: how many cores shall be used (-1 means all) (n_jobs > 1 does not have any effect when 'solver' is set to 'liblinear)
            random_state: random_state for model
            verbose: log level (higher number --> more logs)
            warm_start: work with previous fit and add more estimator
            tol: Tolerance for stopping criteria
            C: Inverse of regularization strength
            max_iter: Maximum number of iterations taken for the solvers to converge

            solver: Algorithm to use in the optimization problem
            penalty: Specify the norm of the penalty
        """
        model_type = "LR"
        model = LogisticRegression(
            random_state=random_state,
            **kwargs,
        )
        grid = ConfigurationSpace(
            seed=42,
            space={
            "solver": Categorical("solver", ["newton-cg", "lbfgs", "liblinear", "sag", "saga"], weights=[0.15, 0.15, 0.15, 0.15, 0.4], default="lbfgs"),
            "penalty": Categorical("penalty", ["l2", "elasticnet"], default="l2"),
            "C": Float("C", (0.01, 100), log=True, default=1),
            "l1_ratio": Float("l1_ratio", (0.01, 1), default=0.1),
            })
        solver_and_penalty = ForbiddenAndConjunction(
            ForbiddenEqualsClause(grid["penalty"], "elasticnet"),
            ForbiddenInClause(grid["solver"], ["newton-cg", "lbfgs", "liblinear", "sag"]),
        )
        l1_ratio_cond = EqualsCondition(grid["l1_ratio"], grid["penalty"], "elasticnet")
        grid.add_forbidden_clause(solver_and_penalty)
        grid.add_condition(l1_ratio_cond)

        # workaround for now -> Problems with EqualsCondition (KeyError: 'l1_ratio') (04/07/2023)
        self.smac_grid = ConfigurationSpace(
            seed=42,
            space={
            "solver": Categorical("solver", ["newton-cg", "lbfgs", "liblinear", "sag", "saga"], default="lbfgs"),
            "penalty": Categorical("penalty", ["l2"], default="l2"),
            "C": Float("C", (0.01, 100), log=True, default=1),
            })
        super().__init__(model, model_name, model_type, grid)



from ConfigSpace import Categorical, ConfigurationSpace, Integer
from sklearn.gaussian_process import GaussianProcessClassifier

from sam_ml.config import get_n_jobs

from .main_classifier import Classifier


class GPC(Classifier):
    """ GaussianProcessClassifier Wrapper class """

    def __init__(
        self,
        model_name: str = "GaussianProcessClassifier",
        n_jobs: int = get_n_jobs(),
        random_state: int = 42,
        **kwargs,
    ):
        """
        @param (important one):
            multi_class: specifies how multi-class classification problems are handled
            max_iter_predict: the maximum number of iterations in Newton's method for approximating the posterior during predict
        """
        model_type = "GPC"
        model = GaussianProcessClassifier(
            n_jobs=n_jobs, random_state=random_state, **kwargs,
        )
        grid = ConfigurationSpace(
            seed=42,
            space={
            "multi_class": Categorical("multi_class", ["one_vs_rest", "one_vs_one"], default="one_vs_rest"),
            "max_iter_predict": Integer("max_iter_predict", (1, 1000), log=True, default=100),
            })
        super().__init__(model, model_name, model_type, grid)



from ConfigSpace import Categorical, ConfigurationSpace, Float
from sklearn.neural_network import MLPClassifier

from .main_classifier import Classifier


class MLPC(Classifier):
    """ MLP Classifier Wrapper class """

    def __init__(
        self,
        model_name: str = "MLP Classifier",
        random_state: int = 42,
        **kwargs,
    ):
        """
        @param (important one):
            hidden_layer_sizes: the ith element represents the number of neurons in the ith hidden layer
            activation: activation function for the hidden layer
            solver: solver for weight optimization
            alpha: l2 penalty (regularization term) parameter
            learning_rate: learning rate schedule for weight updates
            warm_start: work with previous fit and add more estimator
            tol: Tolerance for stopping criteria
            max_iter: Maximum number of iterations taken for the solvers to converge

            random_state: random_state for model
            verbose: logging (True/False)
            batch_size: Size of minibatches for stochastic optimizers
            early_stopping: True: tests on 10% of train data and stops if there is for 'n_iter_no_change' no improvement in the metrics
        """
        model_type = "MLPC"
        model = MLPClassifier(
            random_state=random_state,
            **kwargs,
        )
        grid = ConfigurationSpace(
            seed=42,
            space={
            "hidden_layer_sizes": Categorical("hidden_layer_sizes", ((10, 30, 10), (20,), (10,), (100,), (50,50,50), (50,100,50)), default=(100, )),
            "activation": Categorical("activation", ["tanh", "relu", "logistic"], default="relu"),
            "solver": Categorical("solver", ["sgd", "adam"], default="adam"),
            "alpha": Float("alpha", (0.0001, 0.05), log=True, default=0.0001),
            "learning_rate": Categorical("learning_rate", ["constant", "adaptive"], default="constant"),
            })
        super().__init__(model, model_name, model_type, grid)



from ConfigSpace import Categorical, ConfigurationSpace, Integer
from sklearn.neighbors import KNeighborsClassifier

from .main_classifier import Classifier


class KNC(Classifier):
    """ KNeighborsClassifier Wrapper class """

    def __init__(
        self,
        model_name: str = "KNeighborsClassifier",
        **kwargs,
    ):
        """
        @param (important one):
            n_neighbors: Number of neighbors to use by default for kneighbors queries
            weights: Weight function used in prediction
            algorithm: Algorithm used to compute the nearest neighbors
            leaf_size: Leaf size passed to BallTree or KDTree
            p: number of metric that is used (manhattan, euclidean, minkowski)
            n_jobs: the number of parallel jobs to run for neighbors search [problem with n_jobs = -1 --> kernel dies]
        """
        model_type = "KNC"
        model = KNeighborsClassifier(**kwargs,)
        grid = ConfigurationSpace(
            seed=42,
            space={
            "n_neighbors": Integer("n_neighbors", (1, 30), default=5),
            "p": Integer("p", (1, 5), default=2),
            "leaf_size": Integer("leaf_size", (1, 50), default=30),
            "weights": Categorical("weights", ["uniform", "distance"], default="uniform"),
            })
        super().__init__(model, model_name, model_type, grid)



import warnings

from ConfigSpace import ConfigurationSpace, Float
from sklearn.discriminant_analysis import QuadraticDiscriminantAnalysis

from .main_classifier import Classifier

warnings.filterwarnings("ignore", category=UserWarning)


class QDA(Classifier):
    """ QuadraticDiscriminantAnalysis Wrapper class """

    def __init__(
        self,
        model_name: str = "QuadraticDiscriminantAnalysis",
        **kwargs,
    ):
        """
        @param (important one):
            reg_param: regularizes the per-class covariance estimates by transforming
        """
        model_type = "QDA"
        model = QuadraticDiscriminantAnalysis(**kwargs)
        grid = ConfigurationSpace(
            seed=42,
            space={
            "reg_param": Float("reg_param", (0, 1), default=0),
            })
        super().__init__(model, model_name, model_type, grid)



from ConfigSpace import (
    Categorical,
    ConfigurationSpace,
    Float,
    ForbiddenAndConjunction,
    ForbiddenEqualsClause,
)
from sklearn.svm import LinearSVC

from .main_classifier import Classifier


class LSVC(Classifier):
    """ LinearSupportVectorClassifier Wrapper class """

    def __init__(
        self,
        model_name: str = "LinearSupportVectorClassifier",
        random_state: int = 42,
        **kwargs,
    ):
        """
        @param (important one):
            random_state: random_state for model
            verbose: logging
            penalty: specifies the norm used in the penalization
            dual: select the algorithm to either solve the dual or primal optimization problem
            C: Inverse of regularization strength
            max_iter: Maximum number of iterations taken for the solvers to converge
        """
        model_type = "LSVC"
        model = LinearSVC(
            random_state=random_state,
            **kwargs,
        )
        grid = ConfigurationSpace(
            seed=42,
            space={
            "penalty": Categorical("penalty", ["l1", "l2"], default="l2"),
            "dual": Categorical("dual", [True, False], default=True),
            "C": Float("C", (0.1, 1000), log=True, default=1),
            })
        penalty_dual = ForbiddenAndConjunction(
            ForbiddenEqualsClause(grid["dual"], True),
            ForbiddenEqualsClause(grid["penalty"], "l1"),
        )
        grid.add_forbidden_clause(penalty_dual)
        super().__init__(model, model_name, model_type, grid)



import pandas as pd
from sklearn.preprocessing import (
    MaxAbsScaler,
    MinMaxScaler,
    Normalizer,
    PowerTransformer,
    QuantileTransformer,
    RobustScaler,
    StandardScaler,
)

from sam_ml.config import setup_logger

logger = setup_logger(__name__)


class Scaler:
    """ Scaler Wrapper class """

    def __init__(self, scaler: str = "standard", **kwargs):
        """
        @param:
            scaler: kind of scaler to use
                'standard': StandardScaler
                'minmax': MinMaxScaler
                'maxabs': MaxAbsScaler
                'robust': RobustScaler
                'normalizer': Normalizer
                'power': PowerTransformer with method="yeo-johnson"
                'quantile': QuantileTransformer (default of QuantileTransformer)
                'quantile_normal': QuantileTransformer with output_distribution="normal" (gaussian pdf)

            **kwargs:
                additional parameters for scaler
        """
        self.scaler_type = scaler
        self._grid: dict[str, list] = {} # for pipeline structure

        if scaler == "standard":
            logger.debug("using StandardScaler as scaler")
            self.scaler = StandardScaler(**kwargs)

        elif scaler == "minmax":
            logger.debug("using MinMaxScaler as scaler")
            self.scaler = MinMaxScaler(**kwargs)

        elif scaler == "maxabs":
            logger.debug("using MaxAbsScaler as scaler")
            self.scaler = MaxAbsScaler(**kwargs)

        elif scaler == "robust":
            logger.debug("using RobustScaler as scaler")
            self.scaler = RobustScaler(**kwargs)
            
        elif scaler == "normalizer":
            logger.debug("using Normalizer as scaler")
            self.scaler = Normalizer(**kwargs)

        elif scaler == "power":
            logger.debug("using PowerTransformer as scaler")
            self.scaler = PowerTransformer(**kwargs)

        elif scaler == "quantile":
            logger.debug("using QuantileTransformer as scaler")
            self.scaler = QuantileTransformer(**kwargs)

        elif scaler == "quantile_normal":
            logger.debug("using QuantileTransformer with output_distribution='normal' as scaler")
            self.scaler = QuantileTransformer(output_distribution="normal", **kwargs)

        else:
            logger.error(f"scaler='{scaler}' is no valid input -> using StandardScaler")
            self.scaler = StandardScaler()
            self.scaler_type = "standard"

    def __repr__(self) -> str:
        scaler_params: str = ""
        param_dict = self.get_params(False)
        for key in param_dict:
            if type(param_dict[key]) == str:
                scaler_params += key+"='"+str(param_dict[key])+"', "
            else:
                scaler_params += key+"="+str(param_dict[key])+", "
        return f"Scaler({scaler_params})"

    @staticmethod
    def params() -> dict:
        """
        @return:
            possible values for the parameters of the Scaler class
        """
        param = {"scaler": ["standard", "minmax", "maxabs", "robust", "normalizer", "power", "quantile", "quantile_normal"]}
        return param

    def get_params(self, deep: bool = True):
        return {"scaler": self.scaler_type} | self.scaler.get_params(deep)

    def set_params(self, **params):
        self.scaler.set_params(**params)
        return self

    def scale(self, data: pd.DataFrame, train_on: bool = True) -> pd.DataFrame:
        """
        @param:
            train_on: if True, the scaler will fit_transform. Otherwise just transform

        @return:
            Dataframe with scaled data
        """
        columns = data.columns
        logger.debug("scaling - started")

        if train_on:
            scaled_ar = self.scaler.fit_transform(data)
        else:
            scaled_ar = self.scaler.transform(data)

        scaled_df = pd.DataFrame(scaled_ar, columns=columns)

        logger.debug("scaling - finished")

        return scaled_df

        


import random

import pandas as pd
from sklearn.datasets import make_classification


def rescale(col: pd.Series, range_low: int, range_high: int) -> pd.Series:
    """
    Function to rescale values between any given range
    """
    min_val = min(col)
    max_val = max(col)
    scaled = (range_high - range_low) * ((col-min_val)/(max_val - min_val)) + range_low
    return scaled.apply(round)

def synt_data(columns: dict[str, list], label_name: str = "label", sample_num: int = 1000, classes_num: int = 2, weights: list[float] = None, class_sep: float = 0.98, random_seed: int = 42) -> pd.DataFrame:
    """
    @param:
        columns -> dict {"column_name1": [informativ: bool, range: tuple]}
            inforamtiv: is the feature important?
            range: (min_value, max_value) of column
    """

    total_feature_num = len(columns)
    informativ_features = [i for i in list(columns.keys()) if columns[i][0]==True]
    random_features = [i for i in list(columns.keys()) if columns[i][0]==False]

    X, y = make_classification(n_samples=sample_num, n_features=total_feature_num, n_informative = len(informativ_features), n_redundant=0, n_repeated=0,
                            class_sep=class_sep, random_state=random_seed, weights = weights, n_classes=classes_num)

    # Adding column names
    X = pd.DataFrame(X, columns =informativ_features+random_features)

    # Rescaling features to a believable range. For instance, age between 25 and 55 
    for i in informativ_features:
        X[i] = rescale(X[i],columns[i][1][0], columns[i][1][1])

    for i in random_features:
        random.seed(random_seed)
        X[i] = [random.randint(columns[i][1][0], columns[i][1][1]) for _ in range(sample_num)]

    data = pd.concat([pd.DataFrame(X),pd.Series(y, name=label_name)], axis=1)

    #data.to_csv("synthetic_attrition_dataset.csv", index=False)
    return data



from ConfigSpace import Categorical, ConfigurationSpace, Float, Integer, Normal
from sklearn.ensemble import GradientBoostingClassifier

from .main_classifier import Classifier


class GBM(Classifier):
    """ GradientBoostingMachine Wrapper class """

    def __init__(
        self,
        model_name: str = "GradientBoostingMachine",
        random_state: int = 42,
        **kwargs,
    ):
        """
        @param (important one):
            n_estimator: number of boosting stages to perform
            criterion: function to measure the quality of a split
            max_depth: Maximum number of levels in tree
            min_samples_split: Minimum number of samples required to split a node
            min_samples_leaf: Minimum number of samples required at each leaf node
            max_features: number of features to consider when looking for the best split
            subsample: fraction of samples to be used for fitting the individual base learners
            loss: The loss function to be optimized. 'deviance' refers to deviance (= logistic regression) for classification with probabilistic outputs. For loss 'exponential' gradient boosting recovers the AdaBoost algorithm
            learning_rate: shrinks the contribution of each tree by learning rate

            warm_start: work with previous fit and add more estimator
            random_state: random_state for model
        """
        model_type = "GBM"
        model = GradientBoostingClassifier(random_state=random_state, **kwargs,)
        grid = ConfigurationSpace(
            seed=42,
            space={
            "n_estimators": Integer("n_estimators", (20, 1500), log=True, default=100),
            "max_depth": Integer("max_depth", (1, 15), distribution=Normal(5, 3), default=3),
            "min_samples_split": Integer("min_samples_split", (2, 100), log=True, default=2),
            "min_samples_leaf": Integer("min_samples_leaf", (1, 100), log=True, default=1),
            "max_features": Categorical("max_features", ["auto", "sqrt", "log2"], default="auto"),
            "subsample": Float("subsample", (0.7, 1), default=1),
            "criterion": Categorical("criterion", ["friedman_mse", "squared_error"], default="friedman_mse"),
            "loss": Categorical("loss", ["log_loss", "deviance"], default="log_loss"),
            "learning_rate": Float("learning_rate", (0.005, 0.3), log=True, default=0.1),
            })
        
        # workaround for now -> Problems with Normal distribution (in smac_search) (04/07/2023)
        self.smac_grid = ConfigurationSpace(
            seed=42,
            space={
            "n_estimators": Integer("n_estimators", (20, 1500), log=True, default=100),
            "max_depth": Integer("max_depth", (1, 15), default=3),
            "min_samples_split": Integer("min_samples_split", (2, 100), log=True, default=2),
            "min_samples_leaf": Integer("min_samples_leaf", (1, 100), log=True, default=1),
            "max_features": Categorical("max_features", ["auto", "sqrt", "log2"], default="auto"),
            "subsample": Float("subsample", (0.7, 1), default=1),
            "criterion": Categorical("criterion", ["friedman_mse", "squared_error"], default="friedman_mse"),
            "loss": Categorical("loss", ["log_loss", "deviance"], default="log_loss"),
            "learning_rate": Float("learning_rate", (0.005, 0.3), log=True, default=0.1),
            })
        super().__init__(model, model_name, model_type, grid)



import math

from sklearn.metrics import precision_score, recall_score


def samuel_function(x: float) -> float:
    return math.sqrt(1/(1 + math.e**(12*(0.5-x))))

def lewis_function(x: float) -> float:
    return 1-(0.5-0.5*math.cos((x-1)*math.pi))**4


def s_scoring(y_true: list, y_pred: list, scoring: str = None, pos_label: int = -1, strength: int = 2, score_func = samuel_function) -> float:
    """
    @param:
        y_true, y_pred: data to evaluate on
        
        scoring:
            None: no preference between precision and recall
            'precision': take precision more into account
            'recall': take recall more into account
        
        pos_label:
            pos_label > 0: take <scoring> in class <pos_label> more into account
            pos_label = -1: handle all classes the same

        strength: higher strength means a higher weight for the preferred scoring/pos_label

        score_func: function to use for scoring (default: samuel_function)

    @return:
        score as float between 0 and 1
    """
    prec = precision_score(y_true, y_pred, average=None)
    rec = recall_score(y_true, y_pred, average=None)

    score = 1.0
    for i in range(len(prec)):
        if (scoring=='precision' and pos_label==i) or (scoring=='precision' and type(pos_label)==int and pos_label<=0) or (scoring==None and pos_label==i):
            score *= score_func(prec[i])**strength
        else:
            score *= score_func(prec[i])
    for i in range(len(rec)):
        if (scoring=='recall' and pos_label==i) or (scoring=='recall' and type(pos_label)==int and pos_label<=0) or (scoring==None and pos_label==i):
            score *= score_func(rec[i])**strength
        else:
            score *= score_func(rec[i])

    return score

def l_scoring(y_true: list, y_pred: list, scoring: str = None, pos_label: int = -1, strength: int = 2, score_func = lewis_function) -> float:
    return s_scoring(y_true, y_pred, scoring, pos_label, strength, score_func)



from ConfigSpace import ConfigurationSpace, Float, Integer, Normal
from xgboost import XGBClassifier

from sam_ml.config import get_n_jobs

from .main_classifier import Classifier


class XGBC(Classifier):
    """ SupportVectorClassifier Wrapper class """

    def __init__(
        self,
        model_name: str = "XGBClassifier",
        n_jobs: str = get_n_jobs(),
        random_state: int = 42,
        **kwargs,
    ):
        """
        @param (important one):
            random_state: random_state for model
            n_jobs: how many cores shall be used (-1 means all)
        """
        model_type = "XGBC"
        model = XGBClassifier(
            n_jobs=n_jobs,
            random_state=random_state,
            **kwargs,
        )
        grid = ConfigurationSpace(
            seed=42,
            space={
            "max_depth": Integer("max_depth", (3, 10), default=6),
            "gamma": Float('gamma', (0, 9), default=0),
            'reg_alpha' : Integer('reg_alpha', (0, 180), default=0),
            'reg_lambda' : Float('reg_lambda', (0, 1), default=1),
            'colsample_bytree' : Float('colsample_bytree', (0.5, 1), default=1),
            'min_child_weight' : Integer('min_child_weight', (0, 10), default=1),
            'n_estimators': Integer("n_estimators", bounds=(50, 750), distribution=Normal(150, 100), default=100),
            "learning_rate": Float("learning_rate", bounds=(0.001, 0.30), log=True, default=0.1),
            })
        
        # workaround for now -> Problems with Normal distribution (in smac_search) (04/07/2023)
        self.smac_grid = ConfigurationSpace(
            seed=42,
            space={
            "max_depth": Integer("max_depth", (3, 10), default=6),
            "gamma": Float('gamma', (0, 9), default=0),
            'reg_alpha' : Integer('reg_alpha', (0, 180), default=0),
            'reg_lambda' : Float('reg_lambda', (0, 1), default=1),
            'colsample_bytree' : Float('colsample_bytree', (0.5, 1), default=1),
            'min_child_weight' : Integer('min_child_weight', (0, 10), default=1),
            'n_estimators': Integer("n_estimators", bounds=(50, 750), default=100),
            "learning_rate": Float("learning_rate", bounds=(0.001, 0.30), log=True, default=0.1),
            })
        super().__init__(model, model_name, model_type, grid)

    def feature_importance(self):
        super().feature_importance()



import os


def get_n_jobs() -> int|None:
    n_jobs = os.getenv("SAM_ML_N_JOBS")
    if str(n_jobs) == "-1" or n_jobs is None:
        return -1
    elif str(n_jobs).lower() == "none":
        return None
    elif str(n_jobs).isdigit():
        return int(n_jobs)
    else:
        raise ValueError(f"SAM_ML_N_JOBS cannot be '{n_jobs}' -> has to be 'none', '*positive integer*', or '-1'")
    


from .AdaBoostClassifier import ABC
from .BaggingClassifier import BC
from .BernoulliNB import BNB
from .ClassifierTest import CTest
from .DecisionTreeClassifier import DTC
from .ExtraTreesClassifier import ETC
from .GaussianNB import GNB
from .GaussianProcessClassifier import GPC
from .GradientBoostingMachine import GBM
from .KNeighborsClassifier import KNC
from .LinearDiscriminantAnalysis import LDA
from .LinearSupportVectorClassifier import LSVC
from .LogisticRegression import LR
from .main_classifier import Classifier
from .main_model import Model
from .main_pipeline import Pipeline
from .MLPClassifier import MLPC
from .QuadraticDiscriminantAnalysis import QDA
from .RandomForestClassifier import RFC
from .SupportVectorClassifier import SVC
from .XGBoostClassifier import XGBC

__all__ = {
    "model parent class": "Model",
    "classifier parent class": "Classifier",
    "pipeline class": "Pipeline",
    "AutoML class": "CTest",
    "RandomForestClassifier": "RFC",
    "LogisticRegression": "LR",
    "DecisionTreeClassifier": "DCT",
    "SupportVectorClassifier": "SVC",
    "MLP Classifier": "MLPC",
    "GradientBoostingMachine": "GBM",
    "AdaBoostClassifier": "ABC",
    "KNeighborsClassifier": "KNC",
    "ExtraTreesClassifier": "ETC",
    "GaussianNaiveBayes": "GNB",
    "BernoulliNaiveBayes": "BNB",
    "GaussianProcessClassifier": "GPC",
    "QuadraticDiscriminantAnalysis": "QDA",
    "LinearDiscriminantAnalysis": "LDA",
    "BaggingClassifier": "BC",
    "LinearSupportVectorClassifier": "LSVC",
    "XGBoostClassifier": "XGBC",
}



import numpy as np
from ConfigSpace import Categorical, ConfigurationSpace, Float, InCondition
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis

from .main_classifier import Classifier

np.seterr(divide = 'ignore')

class LDA(Classifier):
    """ LinearDiscriminantAnalysis Wrapper class """

    def __init__(
        self,
        model_name: str = "LinearDiscriminantAnalysis",
        **kwargs,
    ):
        """
        @param (important one):
            solver: solver to use
            shrinkage: shrinkage parameters (does not work with 'svd' solver)
        """
        model_type = "LDA"
        model = LinearDiscriminantAnalysis(**kwargs)
        grid = ConfigurationSpace(
            seed=42,
            space={
            "solver": Categorical("solver", ["lsqr", "eigen", "svd"], weights=[0.475, 0.475, 0.05], default="svd"),
            "shrinkage": Float("shrinkage", (0, 1), default=0),
            })
        shrinkage_cond = InCondition(grid["shrinkage"], grid["solver"], ["lsqr", "eigen"])
        grid.add_condition(shrinkage_cond)
        super().__init__(model, model_name, model_type, grid)



from ConfigSpace import Beta, Categorical, ConfigurationSpace, Float, Integer
from sklearn.base import ClassifierMixin
from sklearn.ensemble import AdaBoostClassifier, RandomForestClassifier
from sklearn.tree import DecisionTreeClassifier

from .main_classifier import Classifier


class ABC(Classifier):
    """ AdaBoostClassifier Wrapper class """

    def __init__(
        self,
        model_name: str = "AdaBoostClassifier",
        random_state: int = 42,
        estimator: ClassifierMixin = DecisionTreeClassifier(max_depth=1),
        **kwargs,
    ):
        """
        @param (important one):
            estimator: base estimator from which the boosted ensemble is built (default: DecisionTreeClassifier with max_depth=1)
            n_estimator: number of boosting stages to perform
            learning_rate: shrinks the contribution of each tree by learning rate
            algorithm: boosting algorithm
            random_state: random_state for model
        """
        model_type = "ABC"
        model = AdaBoostClassifier(
            random_state=random_state,
            estimator=estimator,
            **kwargs,
        )
        
        grid = ConfigurationSpace(
            seed=42,
            space={
            "n_estimators": Integer("n_estimators", (10, 3000), log=True, default=50),
            "learning_rate": Float("learning_rate", (0.005, 2), distribution=Beta(10, 5), default=1),
            "algorithm": Categorical("algorithm", ["SAMME.R", "SAMME"], default="SAMME.R"),
            })
        
        if type(model.estimator) == RandomForestClassifier:
            grid.add_hyperparameter(Integer("estimator__max_depth", (1, 11), default=5))
            grid.add_hyperparameter(Integer("estimator__n_estimators", (5, 100), log=True, default=50))
        elif type(model.estimator) == DecisionTreeClassifier:
            grid.add_hyperparameter(Integer("estimator__max_depth", (1, 11), default=1))
        
        super().__init__(model, model_name, model_type, grid)



from ConfigSpace import ConfigurationSpace, Float
from sklearn.naive_bayes import GaussianNB

from .main_classifier import Classifier


class GNB(Classifier):
    """ GaussianNB Wrapper class """

    def __init__(
        self,
        model_name: str = "GaussianNB",
        **kwargs,
    ):
        """
        @params:
            priors: Prior probabilities of the classes. If specified the priors are not adjusted according to the data
            var_smoothing: Portion of the largest variance of all features that is added to variances for calculation stability
        """
        model_type = "GNB"
        model = GaussianNB(**kwargs,)
        grid = ConfigurationSpace(
            seed=42,
            space={
            "var_smoothing": Float("var_smoothing", (1e-11, 1), log=True, default=1e-9)
            })
        super().__init__(model, model_name, model_type, grid)



import os
import sys
import warnings
from datetime import timedelta
from statistics import mean
from typing import Union

import numpy as np
import pandas as pd
from ConfigSpace import Configuration, ConfigurationSpace
from matplotlib import pyplot as plt
from sklearn.metrics import (
    accuracy_score,
    classification_report,
    make_scorer,
    precision_score,
    recall_score,
)
from sklearn.model_selection import cross_validate
from tqdm.auto import tqdm

from sam_ml.config import get_n_jobs, setup_logger

from .main_model import Model
from .scorer import l_scoring, s_scoring

SMAC_INSTALLED: bool
try:
    from smac import HyperparameterOptimizationFacade, Scenario
    SMAC_INSTALLED = True
except:
    SMAC_INSTALLED = False

logger = setup_logger(__name__)

if not sys.warnoptions:
    warnings.simplefilter("ignore")
    os.environ["PYTHONWARNINGS"] = "ignore" # Also affects subprocesses


class Classifier(Model):
    """ Classifier parent class """

    def __init__(self, model_object = None, model_name: str = "classifier", model_type: str = "Classifier", grid: ConfigurationSpace = ConfigurationSpace()):
        """
        @params:
            model_object: model with 'fit', 'predict', 'set_params', and 'get_params' method (see sklearn API)
            model_name: name of the model
            model_type: kind of estimator (e.g. 'RFC' for RandomForestClassifier)
            grid: hyperparameter grid for the model
        """
        super().__init__(model_object, model_name, model_type)
        self._grid = grid
        self.cv_scores: dict[str, float] = {}
        self.rCVsearch_results: pd.DataFrame|None = None

    def __repr__(self) -> str:
        params: str = ""
        param_dict = self.get_params(False)
        for key in param_dict:
            if type(param_dict[key]) == str:
                params+= key+"='"+str(param_dict[key])+"', "
            else:
                params+= key+"="+str(param_dict[key])+", "
        params += f"model_name='{self.model_name}'"

        return f"{self.model_type}({params})"

    @property
    def grid(self):
        """
        @return:
            hyperparameter tuning grid of the model
        """
        return self._grid
    
    def get_random_config(self):
        """
        @return;
            set of random parameter from grid
        """
        return dict(self.grid.sample_configuration(1))
    
    def get_random_configs(self, n_trails: int) -> list:
        """
        @return;
            n_trails elements in list with sets of random parameterd from grid

        NOTE: filter out duplicates -> could be less than n_trails
        """
        if n_trails<1:
            raise ValueError(f"n_trails has to be greater 0, but {n_trails}<1")
        
        configs = [self._grid.get_default_configuration()]
        if n_trails == 2:
            configs += [self._grid.sample_configuration(n_trails-1)]
        else:
            configs += self._grid.sample_configuration(n_trails-1)
        # remove duplicates
        configs = list(dict.fromkeys(configs))
        return configs

    def replace_grid(self, new_grid: ConfigurationSpace):
        """
        function to replace self.grid 

        e.g.:
            ConfigurationSpace(
                seed=42,
                space={
                    "solver": Categorical("solver", ["lsqr", "eigen", "svd"]),
                    "shrinkage": Float("shrinkage", (0, 1)),
                })
        """
        self._grid = new_grid

    def evaluate(
        self,
        x_test: pd.DataFrame,
        y_test: pd.Series,
        avg: str = None,
        pos_label: Union[int, str] = -1,
        console_out: bool = True,
        secondary_scoring: str = None,
        strength: int = 3,
    ) -> dict[str, float]:
        """
        @param:
            x_test, y_test: Data to evaluate model

            avg: average to use for precision and recall score (e.g. "micro", None, "weighted", "binary")
            pos_label: if avg="binary", pos_label says which class to score. pos_label is used by s_score/l_score

            console_out: shall the result be printed into the console

            secondary_scoring: weights the scoring (only for 's_score'/'l_score')
            strength: higher strength means a higher weight for the preferred secondary_scoring/pos_label (only for 's_score'/'l_score')

        @return: dictionary with keys with scores: 'accuracy', 'precision', 'recall', 's_score', 'l_score'
        """
        pred = self.predict(x_test)

        # Calculate Accuracy, Precision and Recall Metrics
        accuracy = accuracy_score(y_test, pred)
        precision = precision_score(y_test, pred, average=avg, pos_label=pos_label)
        recall = recall_score(y_test, pred, average=avg, pos_label=pos_label)
        s_score = s_scoring(y_test, pred, strength=strength, scoring=secondary_scoring, pos_label=pos_label)
        l_score = l_scoring(y_test, pred, strength=strength, scoring=secondary_scoring, pos_label=pos_label)

        if console_out:
            print("accuracy: ", accuracy)
            print("precision: ", precision)
            print("recall: ", recall)
            print("s_score: ", s_score)
            print("l_score: ", l_score)
            print()
            print("classification report: ")
            print(classification_report(y_test, pred))

        self.test_score = {
            "accuracy": accuracy,
            "precision": precision,
            "recall": recall,
            "s_score": s_score,
            "l_score": l_score,
        }

        return self.test_score
    
    def evaluate_score(
        self,
        x_test: pd.DataFrame,
        y_test: pd.Series,
        scoring: str = "accuracy",
        avg: str = None,
        pos_label: Union[int, str] = -1,
        secondary_scoring: str = None,
        strength: int = 3,
    ) -> float:
        """
        @param:
            x_test, y_test: Data to evaluate model
            scoring: metrics to evaluate the models ("accuracy", "precision", "recall", "s_score", "l_score")

            avg: average to use for precision and recall score (e.g. "micro", None, "weighted", "binary")
            pos_label: if avg="binary", pos_label says which class to score. pos_label is used by s_score/l_score
            secondary_scoring: weights the scoring (only for 's_score'/'l_score')
            strength: higher strength means a higher weight for the preferred secondary_scoring/pos_label (only for 's_score'/'l_score')

        @return: score as float
        """
        pred = self.predict(x_test)

        # Calculate score
        if scoring == "accuracy":
            score = accuracy_score(y_test, pred)
        elif scoring == "precision":
            score = precision_score(y_test, pred, average=avg, pos_label=pos_label)
        elif scoring == "recall":
            score = recall_score(y_test, pred, average=avg, pos_label=pos_label)
        elif scoring == "s_score":
            score = s_scoring(y_test, pred, strength=strength, scoring=secondary_scoring, pos_label=pos_label)
        elif scoring == "l_score":
            score = l_scoring(y_test, pred, strength=strength, scoring=secondary_scoring, pos_label=pos_label)
        else:
            raise ValueError(f"scoring='{scoring}' is not supported -> only  'accuracy', 'precision', 'recall', 's_score', or 'l_score' ")

        return score

    def cross_validation(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        cv_num: int = 10,
        avg: str = "macro",
        pos_label: Union[int, str] = -1,
        console_out: bool = True,
        secondary_scoring: str = None,
        strength: int = 3,
    ) -> dict[str, float]:
        """
        @param:
            X, y: data to cross validate on
            cv_num: number of different splits

            avg: average to use for precision and recall score (e.g. "micro", "weighted", "binary")
            pos_label: if avg="binary", pos_label says which class to score. pos_label is used by s_score/l_score

            console_out: shall the result be printed into the console

            secondary_scoring: weights the scoring (only for 's_score'/'l_score')
            strength: higher strength means a higher weight for the preferred secondary_scoring/pos_label (only for 's_score'/'l_score')

        @return:
            dictionary with "accuracy", "precision", "recall", "s_score", "l_score", train_score", "train_time"
        """
        logger.debug(f"cross validation {self.model_name} - started")

        precision_scorer = make_scorer(precision_score, average=avg, pos_label=pos_label)
        recall_scorer = make_scorer(recall_score, average=avg, pos_label=pos_label)
        s_scorer = make_scorer(s_scoring, strength=strength, scoring=secondary_scoring, pos_label=pos_label)
        l_scorer = make_scorer(l_scoring, strength=strength, scoring=secondary_scoring, pos_label=pos_label)

        if avg == "binary":
            scorer = {
                f"precision ({avg}, label={pos_label})": precision_scorer,
                f"recall ({avg}, label={pos_label})": recall_scorer,
                "accuracy": "accuracy",
                "s_score": s_scorer,
                "l_score": l_scorer,
            }
        else:
            scorer = {
                f"precision ({avg})": precision_scorer,
                f"recall ({avg})": recall_scorer,
                "accuracy": "accuracy",
                "s_score": s_scorer,
                "l_score": l_scorer,
            }

        cv_scores = cross_validate(
            self,
            X,
            y,
            scoring=scorer,
            cv=cv_num,
            return_train_score=True,
            n_jobs=get_n_jobs(),
        )

        pd_scores = pd.DataFrame(cv_scores).transpose()
        pd_scores["average"] = pd_scores.mean(numeric_only=True, axis=1)

        score = pd_scores["average"]
        self.cv_scores = {
            "accuracy": score[list(score.keys())[6]],
            "precision": score[list(score.keys())[2]],
            "recall": score[list(score.keys())[4]],
            "s_score": score[list(score.keys())[8]],
            "l_score": score[list(score.keys())[10]],
            "train_score": score[list(score.keys())[7]],
            "train_time": str(timedelta(seconds = round(score[list(score.keys())[0]]))),
        }

        logger.debug(f"cross validation {self.model_name} - finished")

        if console_out:
            print()
            print(pd_scores)

        return self.cv_scores

    def cross_validation_small_data(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        avg: str = "macro",
        pos_label: Union[int, str] = -1,
        leave_loadbar: bool = True,
        console_out: bool = True,
        secondary_scoring: str = None,
        strength: int = 3,
    ) -> dict[str, float]:
        """
        Cross validation for small datasets (recommended for datasets with less than 150 datapoints)

        @param:
            X, y: data to cross validate on

            avg: average to use for precision and recall score (e.g. "micro", "weighted", "binary")
            pos_label: if avg="binary", pos_label says which class to score. pos_label is used by s_score/l_score

            leave_loadbar: shall the loading bar of the training be visible after training (True - load bar will still be visible)
            console_out: shall the result be printed into the console

            secondary_scoring: weights the scoring (only for 's_score'/'l_score')
            strength: higher strength means a higher weight for the preferred secondary_scoring/pos_label (only for 's_score'/'l_score')

        @return:
            dictionary with "accuracy", "precision", "recall", "s_score", "l_score", train_score", "train_time"
        """
        logger.debug(f"cross validation {self.model_name} - started")

        predictions = []
        true_values = []
        t_scores = []
        t_times = []
        
        for idx in tqdm(X.index, desc=self.model_name, leave=leave_loadbar):
            x_train = X.drop(idx)
            y_train = y.drop(idx)
            x_test = X.loc[[idx]]
            y_test = y.loc[idx]

            train_score, train_time = self.train(x_train, y_train, console_out=False)
            prediction = self.predict(x_test)

            predictions.append(prediction)
            true_values.append(y_test)
            t_scores.append(train_score)
            t_times.append(train_time)

        accuracy = accuracy_score(true_values, predictions)
        precision = precision_score(true_values, predictions, average=avg, pos_label=pos_label)
        recall = recall_score(true_values, predictions, average=avg, pos_label=pos_label)
        s_score = s_scoring(true_values, predictions, strength=strength, scoring=secondary_scoring, pos_label=pos_label)
        l_score = l_scoring(true_values, predictions, strength=strength, scoring=secondary_scoring, pos_label=pos_label)
        avg_train_score = mean(t_scores)
        avg_train_time = str(timedelta(seconds=round(sum(map(lambda f: int(f[0])*3600 + int(f[1])*60 + int(f[2]), map(lambda f: f.split(':'), t_times)))/len(t_times))))

        self.cv_scores = {
            "accuracy": accuracy,
            "precision": precision,
            "recall": recall,
            "s_score": s_score,
            "l_score": l_score,
            "train_score": avg_train_score,
            "train_time": avg_train_time,
        }

        logger.debug(f"cross validation {self.model_name} - finished")

        if console_out:
            print()
            print("classification report:")
            print(classification_report(true_values, predictions))

        return self.cv_scores

    def feature_importance(self) -> plt.show:
        """
        feature_importance() generates a matplotlib plot of the feature importance from self.model
        """
        if not self.trained:
            logger.error("You have to first train the classifier before getting the feature importance")
            return

        if self.model_type == "MLPC":
            importances = [np.mean(i) for i in self.model.coefs_[0]]  # MLP Classifier
        elif self.model_type in ("DTC", "RFC", "GBM", "CBC", "ABC", "ETC", "XGBC"):
            importances = self.model.feature_importances_
        elif self.model_type in ("KNC", "GNB", "BNB", "GPC", "QDA", "BC"):
            logger.warning(f"{self.model_type} does not have a feature importance")
            return
        else:
            importances = self.model.coef_[0]  # "normal"

        feature_importances = pd.Series(importances, index=self.feature_names)

        fig, ax = plt.subplots()
        if self.model_type in ("RFC", "GBM", "ETC"):
            if self.model_type in ("RFC", "ETC"):
                std = np.std(
                    [tree.feature_importances_ for tree in self.model.estimators_], axis=0,
                )
            elif self.model_type == "GBM":
                std = np.std(
                    [tree[0].feature_importances_ for tree in self.model.estimators_], axis=0,
                )
            feature_importances.plot.bar(yerr=std, ax=ax)
        else:
            feature_importances.plot.bar(ax=ax)
        ax.set_title("Feature importances of " + str(self.model_name))
        ax.set_ylabel("use of coefficients as importance scores")
        fig.tight_layout()
        plt.show()
    
    def smac_search(
        self,
        x_train: pd.DataFrame, 
        y_train: pd.Series,
        n_trails: int = 50,
        cv_num: int = 5,
        scoring: str = "accuracy",
        avg: str = "macro",
        pos_label: Union[int, str] = -1,
        secondary_scoring: str = None,
        strength: int = 3,
        small_data_eval: bool = False,
        walltime_limit: float = 600,
        log_level: int = 20,
    ) -> Configuration:
        """
        @params:
            x_train: DataFrame with train features
            y_train: Series with labels

            n_trails: max number of parameter sets to test
            cv_num: number of different splits per crossvalidation (only used when small_data_eval=False)

            scoring: metrics to evaluate the models ("accuracy", "precision", "recall", "s_score", "l_score")
            avg: average to use for precision and recall score (e.g. "micro", "weighted", "binary")
            pos_label: if avg="binary", pos_label says which class to score. Else pos_label is ignored (except scoring='s_score'/'l_score')
            secondary_scoring: weights the scoring (only for scoring='s_score'/'l_score')
            strength: higher strength means a higher weight for the preferred secondary_scoring/pos_label (only for scoring='s_score'/'l_score')

            small_data_eval: if True: trains model on all datapoints except one and does this for all datapoints (recommended for datasets with less than 150 datapoints)
            
            walltime_limit: the maximum time in seconds that SMAC is allowed to run

            log_level: 10 - DEBUG, 20 - INFO, 30 - WARNING, 40 - ERROR, 50 - CRITICAL (SMAC3 library log levels)

        @return: ConfigSpace.Configuration with best hyperparameters (can be used like dict)
        """
        if not SMAC_INSTALLED:
            raise ImportError("SMAC3 library is not installed -> follow instructions in Repo to install SMAC3 (https://github.com/Priapos1004/SAM_ML)")

        logger.debug("starting smac_search")
        # NormalInteger and EqualsCondition in grid are not supported (using workaround for now) (04/07/2023)
        if self.model_type in ("RFC", "ETC", "GBM", "XGBC", "LR"):
            grid = self.smac_grid
        else:
            grid = self.grid

        scenario = Scenario(
            grid,
            n_trials=n_trails,
            deterministic=True,
            walltime_limit=walltime_limit,
        )

        initial_design = HyperparameterOptimizationFacade.get_initial_design(scenario, n_configs=5)
        logger.debug(f"initial_design: {initial_design.select_configurations()}")

        # define target function
        def grid_train(config: Configuration, seed: int) -> float:
            logger.debug(f"config: {config}")
            model = self.get_deepcopy()
            model.set_params(**config)
            if small_data_eval:
                score = model.cross_validation_small_data(x_train, y_train, console_out=False, leave_loadbar=False, avg=avg, pos_label=pos_label, secondary_scoring=secondary_scoring, strength=strength)
            else:
                score = model.cross_validation(x_train, y_train, console_out=False, cv_num=cv_num, avg=avg, pos_label=pos_label, secondary_scoring=secondary_scoring, strength=strength)
            return 1 - score[scoring]  # SMAC always minimizes (the smaller the better)

        # use SMAC to find the best hyperparameters
        smac = HyperparameterOptimizationFacade(
            scenario,
            grid_train,
            initial_design=initial_design,
            overwrite=True,  # If the run exists, we overwrite it; alternatively, we can continue from last state
            logging_level=log_level,
        )

        incumbent = smac.optimize()
        logger.debug("finished smac_search")
        return incumbent

    def randomCVsearch(
        self,
        x_train: pd.DataFrame,
        y_train: pd.Series,
        n_trails: int = 10,
        scoring: str = "accuracy",
        avg: str = "macro",
        pos_label: Union[int, str] = -1,
        secondary_scoring: str = None,
        strength: int = 3,
        small_data_eval: bool = False,
        cv_num: int = 5,
        leave_loadbar: bool = True,
    ) -> tuple[dict, float]:
        """
        @params:
            x_train: DataFrame with train features
            y_train: Series with labels

            n_trails: number of parameter sets to test

            scoring: metrics to evaluate the models ("accuracy", "precision", "recall", "s_score", "l_score")
            avg: average to use for precision and recall score (e.g. "micro", "weighted", "binary")
            pos_label: if avg="binary", pos_label says which class to score. Else pos_label is ignored (except scoring='s_score'/'l_score')
            secondary_scoring: weights the scoring (only for scoring='s_score'/'l_score')
            strength: higher strength means a higher weight for the preferred secondary_scoring/pos_label (only for scoring='s_score'/'l_score')

            small_data_eval: if True: trains model on all datapoints except one and does this for all datapoints (recommended for datasets with less than 150 datapoints)

            cv_num: number of different splits per crossvalidation (only used when small_data_eval=False)

            leave_loadbar: shall the loading bar of the different parameter sets be visible after training (True - load bar will still be visible)

        @return: dictionary with best hyperparameters and float of best_score
        """
        logger.debug("starting randomCVsearch")
        results = []
        configs = self.get_random_configs(n_trails)
        at_least_one_run: bool = False
        try:
            for config in tqdm(configs, desc=f"randomCVsearch ({self.model_name})", leave=leave_loadbar):
                logger.debug(f"config: {config}")
                model = self.get_deepcopy()
                model.set_params(**config)
                if small_data_eval:
                    score = model.cross_validation_small_data(x_train, y_train, console_out=False, leave_loadbar=False, avg=avg, pos_label=pos_label, secondary_scoring=secondary_scoring, strength=strength)
                else:
                    score = model.cross_validation(x_train, y_train, cv_num=cv_num, console_out=False, avg=avg, pos_label=pos_label, secondary_scoring=secondary_scoring, strength=strength)
                config_dict = dict(config)
                config_dict[scoring] = score[scoring]
                results.append(config_dict)
                at_least_one_run = True
        except KeyboardInterrupt:
            logger.info("KeyboardInterrupt - output interim result")
            if not at_least_one_run:
                return {}, -1
            

        self.rCVsearch_results = pd.DataFrame(results, dtype=object).sort_values(by=scoring, ascending=False)

        # for-loop to keep dtypes of columns
        best_hyperparameters = {} 
        for col in self.rCVsearch_results.columns:
            value = self.rCVsearch_results[col].iloc[0]
            if str(value) != "nan":
                best_hyperparameters[col] = value

        best_score = best_hyperparameters[scoring]
        best_hyperparameters.pop(scoring)

        logger.debug("finished randomCVsearch")
        
        return best_hyperparameters, best_score



from ConfigSpace import Categorical, ConfigurationSpace, Integer
from sklearn.naive_bayes import BernoulliNB

from .main_classifier import Classifier


class BNB(Classifier):
    """ BernoulliNB Wrapper class """

    def __init__(
        self,
        model_name: str = "BernoulliNB",
        **kwargs,
    ):
        """
        @params (important one):
            binarize: threshold for binarizing (mapping to booleans) of sample features. If None, input is presumed to already consist of binary vectors
            fit_prior: whether to learn class prior probabilities or not. If false, a uniform prior will be used
        """
        model_type = "BNB"
        model = BernoulliNB(**kwargs,)
        grid = ConfigurationSpace(
            seed=42,
            space={
            "fit_prior": Categorical("fit_prior", [True, False], default=True),
            "binarize": Integer("binarize", (0, 10), default=0),
            })
        super().__init__(model, model_name, model_type, grid)



from ConfigSpace import Categorical, ConfigurationSpace, Integer, Normal
from sklearn.ensemble import ExtraTreesClassifier

from sam_ml.config import get_n_jobs

from .main_classifier import Classifier


class ETC(Classifier):
    """ ExtraTreesClassifier Wrapper class """

    def __init__(
        self,
        model_name: str = "ExtraTreesClassifier",
        n_jobs: int = get_n_jobs(),
        random_state: int = 42,
        **kwargs,
    ):
        """
        @param (important one):
            n_estimators: Number of trees
            max_depth: Maximum number of levels in tree
            n_jobs: how many cores shall be used (-1 means all)
            random_state: random_state for model
            verbose: log level (higher number --> more logs)
            warm_start: work with previous fit and add more estimator

            max_features: Number of features to consider at every split
            min_samples_split: Minimum number of samples required to split a node
            min_samples_leaf: Minimum number of samples required at each leaf node
            bootstrap: Method of selecting samples for training each tree
            criterion: function to measure the quality of a split
        """
        model_type = "ETC"
        model = ExtraTreesClassifier(
            n_jobs=n_jobs,
            random_state=random_state,
            **kwargs,
        )
        grid = ConfigurationSpace(
            seed=42,
            space={
            "n_estimators": Integer("n_estimators", (10, 1000), log=True, default=100),
            "max_depth": Integer("max_depth", (3, 15), distribution=Normal(5, 3), default=5),
            "min_samples_split": Integer("min_samples_split", (2, 10), default=2),
            "min_samples_leaf": Integer("min_samples_leaf", (1, 4), default=1),
            "bootstrap": Categorical("bootstrap", [True, False], default=False),
            "criterion": Categorical("criterion", ["gini", "entropy"], default="gini"),
            })
        
        # workaround for now -> Problems with Normal distribution (in smac_search) (04/07/2023)
        self.smac_grid = ConfigurationSpace(
            seed=42,
            space={
            "n_estimators": Integer("n_estimators", (10, 1000), log=True, default=100),
            "max_depth": Integer("max_depth", (3, 15), default=5),
            "min_samples_split": Integer("min_samples_split", (2, 10), default=2),
            "min_samples_leaf": Integer("min_samples_leaf", (1, 4), default=1),
            "bootstrap": Categorical("bootstrap", [True, False], default=False),
            "criterion": Categorical("criterion", ["gini", "entropy"], default="gini"),
            })
        super().__init__(model, model_name, model_type, grid)



from ConfigSpace import Categorical, ConfigurationSpace, Integer
from sklearn.tree import DecisionTreeClassifier

from .main_classifier import Classifier


class DTC(Classifier):
    """ DecisionTreeClassifier Wrapper class """

    def __init__(
        self,
        model_name: str = "DecisionTreeClassifier",
        random_state: int = 42,
        **kwargs,
    ):
        """
        @param (important one):
            criterion: function to measure the quality of a split
            max_depth: Maximum number of levels in tree
            min_samples_split: Minimum number of samples required to split a node
            min_samples_leaf: Minimum number of samples required at each leaf node
            random_state: random_state for model
        """
        model_type = "DTC"
        model = DecisionTreeClassifier(
            random_state=random_state,
            **kwargs,
        )
        grid = ConfigurationSpace(
            seed=42,
            space={
            "criterion": Categorical("criterion", ["gini", "entropy"], default="gini"),
            "max_depth": Integer("max_depth", (1, 10), default=5),
            "min_samples_split": Integer("min_samples_split", (2, 10), default=2),
            "min_samples_leaf": Integer("min_samples_leaf", (1, 5), default=1),
            })
        super().__init__(model, model_name, model_type, grid)



import logging
import os
import sys


def setup_logger(name=__name__):
    logger = logging.getLogger(name)
    
    log_level = str(os.getenv("SAM_ML_LOG_LEVEL")).lower()
    if log_level == "debug":
        logger.setLevel(logging.DEBUG)
    elif log_level == "info":
        logger.setLevel(logging.INFO)
    elif log_level == "warning":
        logger.setLevel(logging.WARNING)
    elif log_level == "error":
        logger.setLevel(logging.ERROR)
    else:
        os.environ["SAM_ML_LOG_LEVEL"] = "info"
        logger.setLevel(logging.INFO)

    c_handler = logging.StreamHandler(sys.stdout)

    c_format = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    c_handler.setFormatter(c_format)

    # to prevent double logging output
    if logger.hasHandlers():
        logger.handlers.clear()
    logger.propagate = False

    logger.addHandler(c_handler)
    return logger



from typing import Union

import pandas as pd
from imblearn.over_sampling import SMOTE, RandomOverSampler
from imblearn.under_sampling import NearMiss, RandomUnderSampler, TomekLinks
from sklearn.utils import resample

from sam_ml.config import setup_logger

logger = setup_logger(__name__)


def simple_upsample(x_train: pd.DataFrame, y_train: pd.Series, label: Union[int, str] = 1) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    function written by Sughu

    @param:
        x_train, y_train: trainings data for upsamling
        label: label that shall be upsampled
    
    @return:
        tuple x_train, y_train
    """
    count_per_sample = max(y_train.value_counts())
    # Reset indexes
    x_train.reset_index(drop=True)
    y_train.reset_index(drop=True)
    y_train = pd.DataFrame(y_train)
    y_train.columns = ["y"]

    # Stacking horizontally
    df = pd.concat([x_train, y_train], axis=1)

    df_1 = df[df["y"] == label]
    # set other classes to another dataframe
    other_df = df[df["y"] != label]
    
    # upsample the minority class
    df_1_upsampled = resample(
        df_1, random_state=42, n_samples=count_per_sample, replace=True
    )
    
    # concatenate the upsampld dataframe
    df_1_upsampled = df_1_upsampled.reset_index(drop=True)
    other_df = other_df.reset_index(drop=True)

    df_upsampled = pd.concat([df_1_upsampled, other_df])

    # Split into X and y
    x_train = df_upsampled.iloc[:, :-1]
    y_train = df_upsampled.iloc[:, -1]

    return x_train, y_train

class Sampler:
    """ sample algorithm Wrapper class """

    def __init__(self, algorithm: str = "ros", random_state: int = 42, **kwargs):
        """
        @param:
            algorithm: which sampling algorithm to use:
                SMOTE: Synthetic Minority Oversampling Technique (upsampling)
                rus: RandomUnderSampler (downsampling)
                ros: RandomOverSampler (upsampling) (default)
                tl: TomekLinks (downsampling)
                nm: NearMiss (downsampling)
            
            random_state: seed for Random...Sampler

            **kwargs:
                additional parameters for sampler
        """
        self.algorithm = algorithm
        self._grid: dict[str, list] = {} # for pipeline structure

        if algorithm == "SMOTE":
            self.sampler = SMOTE(random_state=random_state, **kwargs)
        elif algorithm == "rus":
            self.sampler = RandomUnderSampler(random_state=random_state, replacement=True, **kwargs)
        elif algorithm == "ros":
            self.sampler = RandomOverSampler(random_state=random_state, **kwargs)
        elif algorithm == "tl":
            self.sampler = TomekLinks(sampling_strategy="majority", **kwargs)
        elif algorithm == "nm":
            self.sampler = NearMiss(**kwargs)
        else:
            logger.error(f"type='{algorithm}' does not exist --> using RandomOverSampler")
            self.sampler = RandomOverSampler(random_state=random_state, **kwargs)
            self.algorithm = "ros"

    def __repr__(self) -> str:
        sampler_params: str = ""
        param_dict = self.get_params(False)
        for key in param_dict:
            if type(param_dict[key]) == str:
                sampler_params += key+"='"+str(param_dict[key])+"', "
            else:
                sampler_params += key+"="+str(param_dict[key])+", "
        return f"Sampler({sampler_params})"

    @staticmethod
    def params() -> dict:
        """
        @return:
            possible values for the parameters
        """
        param = {"algorithm": ["SMOTE", "rus", "ros", "tl", "nm"]}
        return param

    def get_params(self, deep: bool = True):
        return {"algorithm": self.algorithm} | self.sampler.get_params(deep)

    def set_params(self, **params):
        self.sampler.set_params(**params)
        return self

    def sample(self, x_train: pd.DataFrame, y_train: pd.Series) -> tuple[pd.DataFrame, pd.Series]:
        """
        Function for up- and downsampling

        @return:
            tuple x_train_sampled, y_train_sampled
        """
        x_train_sampled, y_train_sampled = self.sampler.fit_resample(x_train, y_train)

        return x_train_sampled, y_train_sampled



from .embeddings import Embeddings_builder
from .feature_selection import Selector
from .sampling import Sampler, simple_upsample
from .scaler import Scaler
from .synthetic_data import synt_data

__all__ = {
    "Synthesis data function": "synt_data",
    "Build embeddings for text": "Embeddings_builder",
    "resample upsampling function": "simple_upsample",
    "imblearn up/downsampling": "Sampler",
    "Scaler class": "Scaler",
    "feature selection class": "Selector"
}



import warnings

from ConfigSpace import Beta, Categorical, ConfigurationSpace, Float, Integer
from sklearn.base import ClassifierMixin
from sklearn.ensemble import BaggingClassifier, RandomForestClassifier
from sklearn.tree import DecisionTreeClassifier

from sam_ml.config import get_n_jobs

from .main_classifier import Classifier

warnings. filterwarnings('ignore')


class BC(Classifier):
    """ BaggingClassifier Wrapper class """

    def __init__(
        self,
        model_name: str = "BaggingClassifier",
        random_state: int = 42,
        n_jobs: int = get_n_jobs(),
        estimator: ClassifierMixin = DecisionTreeClassifier(max_depth=1),
        **kwargs,
    ):
        """
        @param (important one):
            estimator: base estimator from which the boosted ensemble is built (default: DecisionTreeClassifier with max_depth=1)
            n_estimator: number of boosting stages to perform
            max_samples: the number of samples to draw from X to train each base estimator
            max_features: the number of features to draw from X to train each base estimator
            bootstrap: whether samples are drawn with replacement. If False, sampling without replacement is performed
            bootstrap_features: whether features are drawn with replacement
        """
        model_type = "BC"
        model = BaggingClassifier(
            random_state=random_state,
            n_jobs=n_jobs,
            estimator=estimator,
            **kwargs,
        )

        grid = ConfigurationSpace(
            seed=42,
            space={
            "n_estimators": Integer("n_estimators", (3, 3000), distribution=Beta(1, 15), default=10),
            "max_samples": Float("max_samples", (0.1, 1), default=1),
            "max_features": Categorical("max_features", [0.5, 0.9, 1.0, 2, 4], default=1.0),
            "bootstrap": Categorical("bootstrap", [True, False], default=True),
            "bootstrap_features": Categorical("bootstrap_features", [True, False], default=False),
            })
        
        if type(model.estimator) == RandomForestClassifier:
            grid.add_hyperparameter(Integer("estimator__max_depth", (1, 11), default=5))
            grid.add_hyperparameter(Integer("estimator__n_estimators", (5, 100), log=True, default=50))
        elif type(model.estimator) == DecisionTreeClassifier:
            grid.add_hyperparameter(Integer("estimator__max_depth", (1, 11), default=1))
        
        super().__init__(model, model_name, model_type, grid)



import pandas as pd
import statsmodels.api as sm
from sklearn.decomposition import PCA
from sklearn.ensemble import ExtraTreesClassifier
from sklearn.feature_selection import (
    RFE,
    RFECV,
    SelectFromModel,
    SelectKBest,
    SequentialFeatureSelector,
    chi2,
)
from sklearn.linear_model import LogisticRegression
from sklearn.svm import LinearSVC

from sam_ml.config import setup_logger

logger = setup_logger(__name__)


class Selector:
    """ feature selection algorithm Wrapper class """

    def __init__(self, algorithm: str = "kbest", num_features: int = 10, estimator = LinearSVC(penalty="l1", dual=False), **kwargs):
        """
        @params:
            algorithm:
                'kbest': SelectKBest
                'kbest_chi2': SelectKBest with score_func=chi2 (only non-negative values)
                'pca': PCA (new column names after transformation)
                'wrapper': uses p-values of Ordinary Linear Model from statsmodels library (no num_features parameter -> problems with too many features)
                'sequential': SequentialFeatureSelector
                'select_model': SelectFromModel (meta-transformer for selecting features based on importance weights)
                'rfe': RFE (recursive feature elimination)
                'rfecv': RFECV (recursive feature elimination with cross-validation)
            
            estimator:
                parameter is needed for SequentialFeatureSelector, SelectFromModel, RFE, RFECV (default: LinearSVC)
            
            **kwargs:
                additional parameters for selector
        """
        self.algorithm = algorithm
        self.num_features = num_features
        self._grid: dict[str, list] = {} # for pipeline structure

        if algorithm == "kbest":
            self.selector = SelectKBest(k=num_features, **kwargs)
        elif algorithm == "kbest_chi2":
            self.selector = SelectKBest(k=num_features, score_func=chi2, **kwargs)
        elif algorithm == "pca":
            self.selector = PCA(n_components=num_features, random_state=42, **kwargs)
        elif algorithm == "wrapper":
            self.selector = {"pvalue_limit": 0.5}
        elif algorithm == "sequential":
            self.selector = SequentialFeatureSelector(estimator, n_features_to_select=num_features, **kwargs)
        elif algorithm == "select_model":
            self.selector = SelectFromModel(estimator, max_features=num_features, **kwargs)
        elif algorithm == "rfe":
            self.selector = RFE(estimator, n_features_to_select=num_features, **kwargs)
        elif algorithm == "rfecv":
            self.selector = RFECV(estimator, min_features_to_select=num_features, **kwargs)
        else:
            logger.error(f"algorithm='{algorithm}' does not exist -> using SelectKBest algorithm instead")
            self.selector = SelectKBest(k=num_features)
            self.algorithm = "kbest"

    def __repr__(self) -> str:
        selector_params: str = ""
        param_dict = self.get_params(False)
        for key in param_dict:
            if type(param_dict[key]) == str:
                selector_params += key+"='"+str(param_dict[key])+"', "
            else:
                selector_params += key+"="+str(param_dict[key])+", "
        return f"Selector({selector_params})"

    @staticmethod
    def params() -> dict:
        """
        @return:
            possible/recommended values for the parameters
        """
        param = {
            "algorithm": ["kbest", "kbest_chi2", "pca", "wrapper", "sequential", "select_model", "rfe", "rfecv"], 
            "estimator": [LinearSVC(penalty="l1", dual=False), LogisticRegression(), ExtraTreesClassifier(n_estimators=50)]
        }
        return param

    def get_params(self, deep: bool = True):
        class_params = {"algorithm": self.algorithm, "num_features": self.num_features}
        if self.algorithm == "wrapper":
            return class_params | self.selector
        else:
            selector_params = self.selector.get_params(deep)
            if self.algorithm in ("kbest", "kbest_chi2"):
                selector_params.pop("k")
            elif self.algorithm in ("pca"):
                selector_params.pop("n_components")
            elif self.algorithm in ("sequential", "rfe"):
                selector_params.pop("n_features_to_select")
            elif self.algorithm in ("select_model"):
                selector_params.pop("max_features")
            elif self.algorithm in ("rfecv"):
                selector_params.pop("min_features_to_select")

            return class_params | selector_params

    def set_params(self, **params):
        if self.algorithm == "wrapper":
            self.selector = params
        else:
            self.selector.set_params(**params)
        return self
    
    def select(self, X: pd.DataFrame, y: pd.DataFrame = None, train_on: bool = True) -> pd.DataFrame:
        """
        for training: the y data is also needed
        """
        if len(X.columns) < self.num_features:
            logger.warning("the number of features that shall be selected is greater than the number of features in X --> return X")
            self.selected_features = X.columns
            return X

        logger.debug("selecting features - started")
        if train_on:
            if self.algorithm == "wrapper":
                self.selected_features = self.__wrapper_select(X, y, **self.selector)
            else:
                self.selector.fit(X.values, y)
                self.selected_features = self.selector.get_feature_names_out(X.columns)
        
        if self.algorithm == "wrapper":
            X_selected = X[self.selected_features]
        else:
            X_selected = pd.DataFrame(self.selector.transform(X), columns=self.selected_features)

        logger.debug("selecting features - finished")
        return X_selected

    def __wrapper_select(self, X: pd.DataFrame, y: pd.DataFrame, pvalue_limit: float = 0.5, **kwargs) -> list:
        selected_features = list(X.columns)
        y = list(y)
        pmax = 1
        while selected_features:
            p= []
            X_new = X[selected_features]
            X_new = sm.add_constant(X_new)
            model = sm.OLS(y,X_new).fit()
            p = pd.Series(model.pvalues.values[1:],index = selected_features)      
            pmax = max(p)
            feature_pmax = p.idxmax()
            if(pmax>pvalue_limit):
                selected_features.remove(feature_pmax)
            else:
                break
        if len(selected_features) == len(X.columns):
            logger.warning("the wrapper algorithm selected all features")
        return selected_features



import pickle
import time
from copy import deepcopy
from datetime import timedelta
from typing import Union

import pandas as pd

from sam_ml.config import setup_logger

logger = setup_logger(__name__)

class Model:
    """ Model parent class """

    def __init__(self, model_object = None, model_name: str = "model", model_type: str = "Model"):
        """
        @params:
            model_object: model with 'fit', 'predict', 'set_params', and 'get_params' method (see sklearn API)
            model_name: name of the model
            model_type: kind of estimator (e.g. 'RFC' for RandomForestClassifier)
        """
        self.model = model_object
        self.model_name = model_name
        self.model_type = model_type
        self.trained: bool = False
        self.train_score: float = None
        self.train_time: str = None
        self.test_score: Union[float, dict[str, float]] = None
        self.feature_names: list = None

    def __repr__(self) -> str:
        return f"Model(model_object={self.model.__str__()}, model_name='{self.model_name}', model_type='{self.model_type}')"

    def train(self, x_train: pd.DataFrame, y_train: pd.Series, console_out: bool = True) -> tuple[float, str]:
        """
        @return:
            tuple of train score and train time
        """
        logger.debug(f"training {self.model_name} - started")

        start_time = time.time()
        self.model.fit(x_train, y_train)
        end_time = time.time()
        self.feature_names = list(x_train.columns)
        self.train_score = self.model.score(x_train, y_train)
        self.train_time = str(timedelta(seconds=int(end_time-start_time)))

        if console_out:
            print("Train score: ", self.train_score, " - Train time: ", self.train_time)
            
        logger.debug(f"training {self.model_name} - finished")

        self.trained = True

        return self.train_score, self.train_time

    def fit(self, x_train: pd.DataFrame, y_train: pd.Series):
        self.model.fit(x_train, y_train)
        return self

    def predict(self, x_test: pd.DataFrame) -> list:
        """
        @return:
            list with predictions
        """
        return list(self.model.predict(x_test))

    def get_params(self, deep: bool = True):
        return self.model.get_params(deep)

    def set_params(self, **params):
        self.model.set_params(**params)
        return self

    def evaluate(self, x_test: pd.DataFrame, y_test: pd.Series, console_out: bool = True) -> float:
        self.test_score = self.model.score(x_test, y_test)
        if console_out:
            print("Test score: ", self.test_score)
        return self.test_score
    
    def get_deepcopy(self):
        """function to create a deepcopy of object"""
        return deepcopy(self)

    def save_model(self, path: str, only_estimator: bool = False):
        """ 
        function to pickle and save the Class object 
        
        @params:
            path: path to save the model with suffix '.pkl'
            only_estimator: if True, only the estimator of the class object will be saved
        """
        logger.debug(f"saving {self.model_name} - started")
        with open(path, "wb") as f:
            if only_estimator:
                pickle.dump(self.model, f)
            else:
                pickle.dump(self, f)
        logger.debug(f"saving {self.model_name} - finished")

    @staticmethod
    def load_model(path: str):
        """ function to load a pickled model class object """
        logger.debug("loading model - started")
        with open(path, "rb") as f:
            model = pickle.load(f)
        logger.debug("loading model - finished")
        return model



from pathlib import Path

from setuptools import find_packages, setup

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="sam_ml-py",
    version="0.8.0",
    description="a library for ML programing created by Samuel Brinkmann",
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.10",
    packages=find_packages(),
    package_data={},
    scripts=[],
    install_requires=[
        "scikit-learn<1.3", # version 1.3 has some issues currently (04/07/2023)
        "pandas",
        "matplotlib",
        "numpy",
        "imbalanced-learn",
        "pygame",
        "ipywidgets",
        "tqdm",
        "statsmodels",
        "sentence-transformers",
        "xgboost",
        "ConfigSpace", # for hyperparameter tuning spaces
    ],
    extras_require={"test": ["pytest", "pylint", "isort", "refurb", "black"],
                    "with_swig": ["smac"]},
    author="Samuel Brinkmann",
    license="MIT",
    tests_require=["pytest"],
    setup_requires=["pytest-runner"],
)



import os
import sys
import warnings
from typing import Union

import pandas as pd

# to deactivate pygame promt 
os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = '1'

import pygame
from pkg_resources import resource_filename
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from tqdm.auto import tqdm

from sam_ml.config import setup_logger
from sam_ml.data import Embeddings_builder, Sampler, Scaler, Selector

from .AdaBoostClassifier import ABC
from .BaggingClassifier import BC
from .BernoulliNB import BNB
from .DecisionTreeClassifier import DTC
from .ExtraTreesClassifier import ETC
from .GaussianNB import GNB
from .GaussianProcessClassifier import GPC
from .GradientBoostingMachine import GBM
from .KNeighborsClassifier import KNC
from .LinearDiscriminantAnalysis import LDA
from .LinearSupportVectorClassifier import LSVC
from .LogisticRegression import LR
from .main_classifier import Classifier
from .main_pipeline import Pipeline
from .MLPClassifier import MLPC
from .QuadraticDiscriminantAnalysis import QDA
from .RandomForestClassifier import RFC
from .SupportVectorClassifier import SVC
from .XGBoostClassifier import XGBC

logger = setup_logger(__name__)

if not sys.warnoptions:
    warnings.simplefilter("ignore")
    os.environ["PYTHONWARNINGS"] = "ignore" # Also affects subprocesses


class CTest:
    """ AutoML class """

    def __init__(self, models: Union[str, list[Classifier]] = "all", vectorizer: Union[str, Embeddings_builder] = None, scaler: Union[str, Scaler] = None, selector: Union[str, Selector] = None, sampler: Union[str, Sampler] = None):
        """
        @params:
            models:

                - list of Wrapperclass models from this library

                - 'all': use all Wrapperclass models (18+ models)

                - 'basic': use basic Wrapperclass models (8 models) (LogisticRegression, MLP Classifier, LinearSVC, DecisionTreeClassifier, RandomForestClassifier, SVC, Gradientboostingmachine, KNeighborsClassifier)

            vectorizer: type of "data.embeddings.Embeddings_builder" or Embeddings_builder class object for automatic string column vectorizing (None for no vectorizing)
            scaler: type of "data.scaler.Scaler" or Scaler class object for scaling the data (None for no scaling)
            selector: type of "data.feature_selection.Selector" or Selector class object for feature selection (None for no selecting)
            sampling: type of "data.sampling.Sampler" or Sampler class object for sampling the train data (None for no sampling)
        """
        self.__models_input = models

        if type(models) == str:
            models = self.model_combs(models)

        self.models: dict = {}
        for i in range(len(models)):
            self.models[models[i].model_name] = Pipeline(vectorizer,  scaler, selector, sampler, models[i], models[i].model_name+" (pipeline)")

        self._vectorizer = vectorizer
        self._scaler = scaler
        self._selector = selector
        self._sampler = sampler
        self.best_model: Pipeline
        self.scores: dict = {}

    def __repr__(self) -> str:
        params: str = ""

        if type(self.__models_input) == str:
            params += f"models='{self.__models_input}', "
        else:
            params += "models=["
            for model in self.__models_input:
                params += f"\n    {model.__str__()},"
            params += "],\n"

        if type(self._vectorizer) == str:
            params += f"vectorizer='{self._vectorizer}'"
        elif type(self._vectorizer) == Embeddings_builder:
            params += f"vectorizer={self._vectorizer.__str__()}"
        else:
            params += f"vectorizer={self._vectorizer}"
        
        params += ", "

        if type(self._scaler) == str:
            params += f"scaler='{self._scaler}'"
        elif type(self._scaler) == Scaler:
            params += f"scaler={self._scaler.__str__()}"
        else:
            params += f"scaler={self._scaler}"

        params += ", "

        if type(self._selector) == str:
            params += f"selector='{self._selector}'"
        elif type(self._selector) == Selector:
            params += f"selector={self._selector.__str__()}"
        else:
            params += f"selector={self._selector}"

        params += ", "

        if type(self._sampler) == str:
            params += f"sampler='{self._sampler}'"
        elif type(self._sampler) == Sampler:
            params += f"sampler={self._sampler.__str__()}"
        else:
            params += f"sampler={self._sampler}"

        return f"CTest({params})"

    def remove_model(self, model_name: str):
        del self.models[model_name]

    def add_model(self, model: Classifier):
        self.models[model.model_name] = Pipeline(self._vectorizer, self._scaler, self._selector, self._sampler, model, model.model_name+" (pipeline)")

    def model_combs(self, kind: str):
        """
        @params:
            kind:
                "all": use all models
                "basic": use a simple combination (LogisticRegression, MLP Classifier, LinearSVC, DecisionTreeClassifier, RandomForestClassifier, SVC, Gradientboostingmachine, AdaboostClassifier, KNeighborsClassifier)
        """
        if kind == "all":
            models = [
                LR(),
                QDA(),
                LDA(),
                MLPC(),
                LSVC(),
                DTC(),
                RFC(),
                SVC(),
                GBM(),
                
                ABC(model_name="AdaBoostClassifier (DTC based)"),
                ABC(
                    estimator=RandomForestClassifier(max_depth=5, random_state=42),
                    model_name="AdaBoostClassifier (RFC based)",
                ),
                ABC(
                    estimator=LogisticRegression(),
                    model_name="AdaBoostClassifier (LR based)",
                ),
                KNC(),
                ETC(),
                GNB(),
                BNB(),
                GPC(),
                BC(model_name="BaggingClassifier (DTC based)"),
                BC(
                    estimator=RandomForestClassifier(max_depth=5, random_state=42),
                    model_name="BaggingClassifier (RFC based)",
                ),
                BC(
                    estimator=LogisticRegression(),
                    model_name="BaggingClassifier (LR based)",
                ),
                XGBC(),
            ]
        elif kind == "basic":
            models = [
                LR(),
                MLPC(),
                LSVC(),
                DTC(),
                RFC(),
                SVC(),
                GBM(),
                KNC(),
            ]
        else:
            print(f"Cannot find model combination '{kind}' --> using all models")
            models = self.model_combs("all")

        return models

    def __finish_sound(self):
        """ little function to play a microwave sound """
        filepath = resource_filename(__name__, 'microwave_finish_sound.mp3')
        pygame.mixer.init()
        pygame.mixer.music.load(filepath)
        pygame.mixer.music.play()

    def eval_models(
        self,
        x_train: pd.DataFrame,
        y_train: pd.Series,
        x_test: pd.DataFrame,
        y_test: pd.Series,
        avg: str = "macro",
        pos_label: Union[int, str] = -1,
        secondary_scoring: str = None,
        strength: int = 3,
    ) -> dict[str, dict]:
        """
        @param:
            x_train, y_train, x_test, y_test: Data to train and evaluate models

            avg: average to use for precision and recall score (e.g. "micro", "weighted", "binary")    
            pos_label: if avg="binary", pos_label says which class to score. pos_label is used by s_score/l_score

            secondary_scoring: weights the scoring (only for 's_score'/'l_score')
            strength: higher strength means a higher weight for the preferred secondary_scoring/pos_label (only for 's_score'/'l_score')

        @return:
            saves metrics in dict self.scores and also outputs them
        """
        try:
            for key in tqdm(self.models.keys(), desc="Crossvalidation"):
                tscore, ttime = self.models[key].train(x_train, y_train, console_out=False)
                score = self.models[key].evaluate(
                    x_test, y_test, avg=avg, pos_label=pos_label, console_out=False, secondary_scoring=secondary_scoring, strength=strength,
                )
                score["train_score"] = tscore
                score["train_time"] = ttime
                self.scores[key] = score

            self.__finish_sound()
            return self.scores

        except KeyboardInterrupt:
            logger.info("KeyboardInterrupt - output interim result")
            return self.scores

    def eval_models_cv(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        cv_num: int = 5,
        avg: str = "macro",
        pos_label: Union[int, str] = -1,
        small_data_eval: bool = False,
        secondary_scoring: str = None,
        strength: int = 3,
    ) -> dict[str, dict]:
        """
        @param:
            X, y: Data to train and evaluate models on
            cv_num: number of different splits (ignored if small_data_eval=True)

            avg: average to use for precision and recall score (e.g. "micro", "weighted", "binary")
            pos_label: if avg="binary", pos_label says which class to score. pos_label is used by s_score/l_score
            
            small_data_eval: if True: trains model on all datapoints except one and does this for all datapoints (recommended for datasets with less than 150 datapoints)

            secondary_scoring: weights the scoring (only for 's_score'/'l_score')
            strength: higher strength means a higher weight for the preferred secondary_scoring/pos_label (only for 's_score'/'l_score')

        @return:
            saves metrics in dict self.scores and also outputs them
        """

        try:
            for key in tqdm(self.models.keys(), desc="Crossvalidation"):
                if small_data_eval:
                    self.models[key].cross_validation_small_data(
                        X, y, avg=avg, pos_label=pos_label, console_out=False, leave_loadbar=False, secondary_scoring=secondary_scoring, strength=strength,
                    )
                else:
                    self.models[key].cross_validation(
                        X, y, cv_num=cv_num, avg=avg, pos_label=pos_label, console_out=False, secondary_scoring=secondary_scoring, strength=strength,
                    )
                self.scores[key] = self.models[key].cv_scores
            self.__finish_sound()
            return self.scores

        except KeyboardInterrupt:
            logger.info("KeyboardInterrupt - output interim result")
            return self.scores

    def output_scores_as_pd(self, sort_by: Union[str, list[str]] = "index", console_out: bool = True) -> pd.DataFrame:
        """
        @param:
            sorted_by:
                'index': sort index ascending=True
                'precision'/'recall'/'accuracy'/'train_score'/'train_time': sort by these columns ascending=False

                e.g. ['precision', 'recall'] - sort first by 'precision' and then by 'recall'
        """
        if self.scores != {}:
            if sort_by == "index":
                scores = pd.DataFrame.from_dict(self.scores, orient="index").sort_index(ascending=True)
            else:
                scores = (
                    pd.DataFrame.from_dict(self.scores, orient="index")
                    .sort_values(by=sort_by, ascending=False)
                )

            if console_out:
                print(scores)
        else:
            logger.warning("no scores are created -> use 'eval_models()'/'eval_models_cv()' to create scores")
            scores = None

        return scores

    def find_best_model_randomCV(
        self,
        x_train: pd.DataFrame,
        y_train: pd.Series,
        x_test: pd.DataFrame,
        y_test: pd.Series,
        n_trails: int = 5,
        scoring: str = "accuracy",
        avg: str = "macro",
        pos_label: Union[int, str] = -1,
        secondary_scoring: str = None,
        strength: int = 3,
        small_data_eval: bool = False,
        cv_num: int = 3,
        leave_loadbar: bool = True,
    ) -> dict:
        """
        @params:
            x_train: DataFrame with train features
            y_train: Series with train labels
            x_test: DataFrame with test features
            y_test: Series with test labels

            n_trails: number of parameter sets to test

            scoring: metrics to evaluate the models
            avg: average to use for precision and recall score (e.g. "micro", "weighted", "binary")
            pos_label: if avg="binary", pos_label says which class to score. Else pos_label is ignored (except scoring='s_score'/'l_score')
            secondary_scoring: weights the scoring (only for scoring='s_score'/'l_score')
            strength: higher strength means a higher weight for the preferred secondary_scoring/pos_label (only for scoring='s_score'/'l_score')

            small_data_eval: if True: trains model on all datapoints except one and does this for all datapoints (recommended for datasets with less than 150 datapoints)

            cv_num: number of different splits per crossvalidation (only used when small_data_eval=False)

            leave_loadbar: shall the loading bar of the randomCVsearch of each individual model be visible after training (True - load bar will still be visible)
        """
        for key in tqdm(self.models.keys(), desc="randomCVsearch"):
            best_hyperparameters, best_score = self.models[key].randomCVsearch(x_train, y_train, n_trails, scoring, avg, pos_label, secondary_scoring, strength, small_data_eval, cv_num, leave_loadbar)
            logger.info(f"{self.models[key].model_name} - score: {best_score} ({scoring}) - parameters: {best_hyperparameters}")
            if best_hyperparameters != {}:
                model_best = self.models[key].get_deepcopy()
                model_best.set_params(**best_hyperparameters)
                train_score, train_time = model_best.train(x_train, y_train, console_out=False)
                scores = model_best.evaluate(x_test, y_test, avg=avg, pos_label=pos_label, secondary_scoring=secondary_scoring, strength=strength, console_out=False)
                
                scores["train_time"] = train_time
                scores["train_score"] = train_score
                scores["best_score (rCVs)"] = best_score
                scores["best_hyperparameters (rCVs)"] = best_hyperparameters
                self.scores[key] = scores
        sorted_scores = self.output_scores_as_pd(sort_by=[scoring, "s_score", "train_time"], console_out=False)
        best_model_type = sorted_scores.iloc[0].name
        best_model_value = sorted_scores.iloc[0][scoring]
        best_model_hyperparameters = sorted_scores.iloc[0]["best_hyperparameters (rCVs)"]
        logger.info(f"best model type {best_model_type} - {scoring}: {best_model_value} - parameters: {best_model_hyperparameters}")
        self.__finish_sound()
        return self.scores



import pandas as pd
import pytest
from sklearn.datasets import make_classification

from sam_ml.models import (
    ABC,
    BC,
    BNB,
    DTC,
    ETC,
    GBM,
    GNB,
    GPC,
    KNC,
    LDA,
    LR,
    LSVC,
    MLPC,
    QDA,
    RFC,
    SVC,
    XGBC,
    Pipeline,
)
from sam_ml.models.main_classifier import SMAC_INSTALLED

MODELS = [ABC(), BC(), BNB(), DTC(), ETC(), GNB(), GPC(), GBM(), KNC(), LDA(), LSVC(), LR(), MLPC(), QDA(), RFC(), SVC(), XGBC()]
X, Y = make_classification(n_samples = 50,
                            n_features = 5,
                            n_informative = 5,
                            n_redundant = 0,
                            n_classes = 3,
                            weights = [.2, .3, .8])
X = pd.DataFrame(X, columns=["col1", "col2", "col3", "col4", "col5"])
Y = pd.Series(Y)


def test_classifier_train_evaluate():
    for classifier in MODELS:
        classifier.train(X, Y, console_out=False)
        classifier.evaluate(X, Y, console_out=False)
        classifier.evaluate_score(X, Y)


def test_pipelines_train_evaluate():
    for classifier in MODELS:
        model = Pipeline(model=classifier, model_name=classifier.model_name)
        model.train(X, Y, console_out=False)
        model.evaluate(X, Y, console_out=False)
        model.evaluate_score(X, Y)

def test_classifier_crossvalidation():
    for classifier in MODELS:
        classifier.cross_validation(X, Y, cv_num=3)

def test_pipelines_crossvalidation():
    for classifier in MODELS:
        model = Pipeline(model=classifier, model_name=classifier.model_name)
        model.cross_validation(X, Y, cv_num=3)

def test_classifier_crossvalidation_small_data():
    for classifier in MODELS:
        classifier.cross_validation_small_data(X, Y)

def test_pipelines_crossvalidation_small_data():
    for classifier in MODELS:
        model = Pipeline(model=classifier, model_name=classifier.model_name)
        model.cross_validation_small_data(X, Y)

def test_classifier_randomCVsearch():
    for classifier in MODELS:
        best_param, _ = classifier.randomCVsearch(X, Y, n_trails=5, cv_num=3)
        assert best_param != {}, "should always find a parameter combination"

def test_pipelines_randomCVsearch():
    for classifier in MODELS:
        model = Pipeline(model=classifier, model_name=classifier.model_name)
        best_param, _ = model.randomCVsearch(X, Y, n_trails=5, cv_num=3)
        assert best_param != {}, "should always find a parameter combination"

def test_classifier_smac_search():
    if SMAC_INSTALLED:
        for classifier in MODELS:
            best_param = classifier.smac_search(X, Y, n_trails=10, cv_num=3)
            assert best_param != {}, "should always find a parameter combination"
    else:
        with pytest.raises(ImportError):
            for classifier in MODELS:
                best_param = classifier.smac_search(X, Y, n_trails=10, cv_num=3)
                assert best_param != {}, "should always find a parameter combination"

def test_pipelines_smac_search():
    if SMAC_INSTALLED:
        for classifier in MODELS:
            model = Pipeline(model=classifier, model_name=classifier.model_name)
            best_param = model.smac_search(X, Y, n_trails=10, cv_num=3)
            assert best_param != {}, "should always find a parameter combination"
    else:
        with pytest.raises(ImportError):
            for classifier in MODELS:
                model = Pipeline(model=classifier, model_name=classifier.model_name)
                best_param = model.smac_search(X, Y, n_trails=10, cv_num=3)
                assert best_param != {}, "should always find a parameter combination"



import copy
from contextlib import suppress
from typing import Union

import pandas as pd

from sam_ml.config import setup_logger
from sam_ml.data import Embeddings_builder, Sampler, Scaler, Selector

from .main_classifier import Classifier
from .RandomForestClassifier import RFC

logger = setup_logger(__name__)


class Pipeline(Classifier):
    """ classifier pipeline class """

    def __init__(self, vectorizer: Union[str, Embeddings_builder] = None, scaler: Union[str, Scaler] = None, selector: Union[str, Selector] = None, sampler: Union[str, Sampler] = None, model: Union[tuple, Classifier] = RFC(), model_name: str = "pipe"):
        """
        @params:
            vectorizer: type of "data.embeddings.Embeddings_builder" or Embeddings_builder class object for automatic string column vectorizing (None for no vectorizing)
            scaler: type of "data.scaler.Scaler" or Scaler class object for scaling the data (None for no scaling)
            selector: type of "data.feature_selection.Selector" or Selector class object for feature selection (None for no selecting)
            sampling: type of "data.sampling.Sampler" or Sampler class object for sampling the train data (None for no sampling)
            model: Classifier class object or tuple (model, model_type, hyperparameter grid)
            model_name: name of the model
        """
        self._classifier: tuple

        if issubclass(type(model), Classifier):
            with suppress(BaseException):
                self.smac_grid = model.smac_grid
            super().__init__(model.model, model_name, model.model_type, model.grid)
            self._classifier = (model.model, model.model_type, model.grid)
        else:
            super().__init__(model[0], model_name, model[1], model[2])
            self._classifier = model

        if vectorizer in Embeddings_builder.params()["vec"]:
            self.vectorizer = Embeddings_builder(vec=vectorizer)
        elif type(vectorizer) == Embeddings_builder or vectorizer is None:
            self.vectorizer = vectorizer
        else:
            logger.error(f"wrong input '{vectorizer}' for vectorizer -> vectorizer = None")
            self.vectorizer = None

        if scaler in Scaler.params()["scaler"]:
            self.scaler = Scaler(scaler=scaler)
        elif type(scaler) == Scaler or scaler is None:
            self.scaler = scaler
        else:
            logger.error(f"wrong input '{scaler}' for scaler -> scaler = None")
            self.scaler = None

        if selector in Selector.params()["algorithm"]:
            self.selector = Selector(algorithm=selector)
        elif type(selector) == Selector or selector is None:
            self.selector = selector
        else:
            logger.error(f"wrong input '{selector}' for selector -> selector = None")
            self.selector = None

        if sampler in Sampler.params()["algorithm"]:
            self.sampler = Sampler(algorithm=sampler)
        elif type(sampler) == Sampler or sampler is None:
            self.sampler = sampler
        else:
            logger.error(f"wrong input '{sampler}' for sampler -> sampler = None")
            self.sampler = None

        # check for incompatible sampler-model combination
        if self.sampler is not None:
            sampling_problems = ["QDA", "LDA", "LR", "MLPC", "LSVC"]
            if self.sampler.algorithm == "SMOTE" and self.model_type in sampling_problems:
                logger.warning(self.model_type+" does not work with sampling='SMOTE' --> going on with sampling='ros'")
                self.sampler = Sampler(algorithm="ros")
            elif self.sampler.algorithm in ("nm", "tl") and self.model_type in sampling_problems:
                logger.warning(self.model_type+f" does not work with sampling='{self.sampler.algorithm}' --> going on with sampling='rus'")
                self.sampler = Sampler(algorithm="rus")

        self.vectorizer_dict: dict[str, Embeddings_builder] = {}

    def __repr__(self) -> str:
        params: str = ""
        data_steps = ("vectorizer", self.vectorizer), ("scaler", self.scaler), ("selector", self.selector), ("sampler", self.sampler)
        for step in data_steps:
            params += step[0]+"="+step[1].__str__()+", "

        params += f"model={self.model.__str__()}, "
        params += f"model_name='{self.model_name}'"

        return f"Pipeline({params})"

    @property
    def steps(self) -> list[tuple[str, any]]:
        return [("vectorizer", self.vectorizer), ("scaler", self.scaler), ("selector", self.selector), ("sampler", self.sampler), ("model", self._classifier)]
    
    def __auto_vectorizing(self, X: pd.DataFrame, train_on: bool = True) -> pd.DataFrame:
        """ detects string columns, creates a vectorizer for each, and vectorizes them """
        if train_on:
            X = X.convert_dtypes()
            string_columns = list(X.select_dtypes(include="string").columns)
            self._string_columns = string_columns
            self.vectorizer_dict = dict(zip(self._string_columns, [copy.deepcopy(self.vectorizer) for i in range(len(string_columns))]))

        for col in self._string_columns:
            X = pd.concat([X, self.vectorizer_dict[col].vectorize(X[col], train_on=train_on)], axis=1)
        X_vec = X.drop(columns=self._string_columns)

        return X_vec

    def __data_prepare(self, X: pd.DataFrame, y: pd.Series, train_on: bool = True) -> tuple[pd.DataFrame, pd.Series]:
        """ runs data class objects on data to prepare them for the model """
        if self.vectorizer is not None:
            X = self.__auto_vectorizing(X, train_on=train_on)
        if self.scaler is not None:
            X = self.scaler.scale(X, train_on=train_on)
        if self.selector is not None:
            X = self.selector.select(X, y, train_on=train_on)
        if self.sampler is not None and train_on:
            X, y = self.sampler.sample(X, y)
        return X, y

    def train(self, x_train: pd.DataFrame, y_train: pd.Series, console_out: bool = True) -> tuple[float, str]:
        x_train_pre, y_train_pre = self.__data_prepare(x_train, y_train, train_on=True)
        return super().train(x_train_pre, y_train_pre, console_out)

    def fit(self, x_train: pd.DataFrame, y_train: pd.Series):
        x_train_pre, y_train_pre = self.__data_prepare(x_train, y_train, train_on=True)
        return super().fit(x_train_pre, y_train_pre)

    def predict(self, x_test: pd.DataFrame) -> list:
        x_test_pre, _ = self.__data_prepare(x_test, None, train_on=False)
        return super().predict(x_test_pre)

    def get_params(self, deep: bool = True) -> dict[str, any]:
        return dict(self.steps)
